import numpy as np
from scipy.special import sici

mpc2 = .93828E9 # proton mass in eV / c^2
mc2 = 2.*mpc2 # H2 mass
hbarc = 1973.3 # eV ang

muv = .5 # vibrational reduced mass (amu)

mus = 2. # translational mass (amu)

mux = 28. # phonon reduced mass (i.e. Si mass) (amu)

# reaction coord potential params
vd =  2.5 # eV
va = .72 # eV
lmbda = 2.7 # 1/ang

# reaction path curvature params
thetar = 65./180.*np.pi # degrees
lmbdar = 4. # 1/ang
sc = .1 # ang
ro = 0.74144 # gas phase eq H-H bond length (ang)
gamma = 0.5772156649 # euler-mascheroni constant

# phonon params
deltao = .55 # ang
lmbdad = 2.7 # 1/ang
lmbdap = 2.0 # 1/ang
hbaromegapg = .055 # eV
hbaromegaps = .043 # eV
sp = 0. # ang
sd = -.15 # ang

# vibrational params
hbaromegavo = .516 # eV
hbaromegavs = .08 # .08 # eV
lmbdav = 1.6 # 1/ang
sv = .15 # ang

def eta(s,v):
    return 1. - C(s)*v

# curvature of reaction path (in 1/ang)
def C(s):
    return thetar * lmbdar / 2. / np.cosh(lmbdar*(s-sc))**2

# vibrational energy quantum (in eV) as function
def hbaromegav(s):
    return hbaromegavo + (hbaromegavo-hbaromegavs)*(
        np.tanh(lmbdav*(s-sv))-1.
    )

# acceleration of vibrational coord
def fv(s,sdot,v):
    return -v*(hbaromegav(s)/hbaromegavo)**2-C(s)*eta(s,v)*sdot**2

# potential energy of vibrational coord
def vv(s,v):
    return mc2*(hbaromegav(s)/hbarc)**2/2.*v**2

# potential energy of reaction path coord (eV)
def vs(s):
    return (vd-va)*(
        np.tanh(
            lmbda*s
        ) - 1.
    ) / 2. +  1. / 4. * (
        vd + va + 2.*np.sqrt(vd*va)
    ) / np.cosh(lmbda*s)**2

# derivative of vibrational potential
# wrt reaction path coord ( eV / ang )
def dvvds(s,v):
    return mc2 / hbarc**2 * dhbaromegavds(s) * hbaromegav(s) * v**2

# derivative of vibrational frequency
# wrt reaction path coord ( eV / ang )
def dhbaromegavds(s):
    return ( hbaromegavo - hbaromegavs ) * lmbdav / np.cosh(lmbdav * (s - sv))**2

# derivative of reaction path potential
# wrt reaction path coord ( eV / ang )
def dvsds(s):
    return lmbda / 2. / np.cosh(lmbda*s)**2 * (
        ( vd - va ) - (
            vd + va + 2. * np.sqrt(vd*va)
        ) * np.tanh(lmbda*s)
    )

# derivative of reaction path curvature
# wrt reaction path coord ( 1 / ang^2 )
def dcds(s):
    return - thetar * lmbdar**2 / np.cosh(lmbdar*(s-sc))**2 * np.tanh(lmbdar*(s-sc))

# derivative of phonon frequency
# wrt reaction path coord ( eV / ang )
def dhbaromegaxds(s):
    return (hbaromegapg-hbaromegaps)*lmbdap/np.cosh(
        lmbdap*(s-sp)
    )**2

# derivative of phonon equilibrium displacement
# wrt reaction path coord (dimensionless)
def ddeltaxds(s):
    return deltao/2.*lmbdad/np.cosh(
        lmbdad*(s-sd)
    )**2

# derivative of phonon potential energy
# wrt reaction path coord (dimensionless)
def dvxds(s,x):
    return mux/mus*mc2/hbarc**2*hbaromegax(s)*(x-deltax(s))*(
        dhbaromegaxds(s)*(x-deltax(s))-ddeltaxds(s)*hbaromegax(s)
    )

# acceleration of reaction path coord ( ang omegavo^2 )
def fs(s,sdot,v,vdot,x):
    return -1. / eta(s,v)**2 * (
        (dvsds(s)+dvvds(s,v)+dvxds(s,x))/(mc2*hbaromegavo**2/hbarc**2)
    ) + 1. / eta(s,v) * sdot * (
        v * dcds(s) * sdot + 2. * C(s) * vdot
    )

# phonon frequency ( eV )
def hbaromegax(s):
    return hbaromegapg + (
        hbaromegapg - hbaromegaps
    ) * (
        np.tanh(
            lmbdap*(s-sp)
        ) - 1.
    )

# phonon coordinate equilibrium displacement ( ang )
def deltax(s):
    return deltao*(
        1. + np.tanh(lmbdad*(s-sd))
    )/2.

# acceleration of phonon coordinate ( ang omegavo**2 )
def fx(s,x):
    return -(hbaromegax(s)/hbaromegavo)**2*(x-deltax(s))

# phonon coordinate potential energy ( eV )
def vx(s,x):
    return 1./2.*mux/mus*mc2/hbarc**2*hbaromegax(s)**2*(x-deltax(s))**2

# tangent angle of reaction path (radians)
def theta(s):
    return thetar / 2. * (
        1 - np.tanh(lmbdar*(s-sc))
    )

# bond length (r) and
# molecule-surface distance coordinate (z),
# both in units of ang and absolute (i.e.
# not mass-scaled)
# z offset chosen so that z asymptotes
# to s
def rz(s,v):
    thta = theta(s)
    sithr, cithr = sici(thetar)
    simthr, cimthr = sici(-thetar)
    sith, cith = sici(thta)
    siththr, ciththr = sici(thta-thetar)
    return (
        ro + (
            - 1. / 2. / lmbdar * (
                (
                    ciththr - cimthr
                ) * np.sin(thetar) - sith + np.cos(thetar) * (
                    siththr + sithr
                )
            ) + v * np.cos(thta)
        )*np.sqrt(
            mus/muv
        ),
        s + 1. / 2. / lmbdar * (
            -gamma + cith + np.cos(thetar) * (
                cimthr - ciththr
            ) + np.log((thetar-thta)/thta/thetar) + np.sin(thetar) * (
                siththr + sithr
            )
        ) + v * np.sin(thta)
    )

# calculates kinetic and potential energies
# of s and v coordinates
# note:
#       s asymptotes to z and
#       v asymptotes to mass-scaled r
def energies(s,sdot,v,vdot,x,xdot):
    ts = 1./2.*hbaromegavo**2/hbarc**2*mc2*eta(s,v)**2*sdot**2
    tv = 1./2.*hbaromegavo**2/hbarc**2*mc2*vdot**2
    tx = 1./2.*hbaromegavo**2/hbarc**2*mux/mus*mc2*xdot**2
    us = vs(s)
    uv = vv(s,v)
    ux = vx(s,x)
    return (
        ts,tv,tx,us,uv,ux
    )

###########################
# calculate sv trajectory #
#     (using RK4)         #
###########################
#
# returns an 8 element list containing:
# 
# * s coordinates
# * s velocities
# * v coordinates
# * v velocities
# * s kinetic energy
# * v kinetic energy
# * s potential energy
# * v potential energy
#
def calculate_sv_trajectory(
        dt, # time step in vibrational periods (1/omegavo)
        s, # initial s-coord (ang)
        sdot, # initial s-coord velocity (ang . omegavo)
        v, # initial v-coord (ang)
        vdot, # initial v-coord velocity (ang . omegavo)
        smin, smax, # terminate simulation if s < smin or s > smax
        static_surface = False,
):
    S = []
    Sdot = []
    V = []
    Vdot = []
    Ts = []
    Tv = []
    Us = []
    Uv = []
    while s > smin and s < smax:
        ts,tv,tx,us,uv,ux = energies(
            s,sdot,v,vdot,
            deltao if static_surface else deltax(s),
            0
        )
        S.append(s)
        Sdot.append(sdot)
        V.append(v)
        Vdot.append(vdot)
        Ts.append(ts)
        Tv.append(tv)
        Us.append(us+ux)
        Uv.append(uv)
        def k(y):
            return dt * np.array(
                [
                    y[1],
                    fs(
                        y[0],y[1],y[2],y[3],
                        (
                            deltao
                            if static_surface else
                            deltax(y[0])
                        )
                    ),
                    y[3],
                    fv(y[0],y[1],y[2])
                ]
            )
        yo = np.array(
            [s,sdot,v,vdot]
        )
        k1 = k(yo)
        k2 = k(yo+k1/2.)
        k3 = k(yo+k2/2.)
        k4 = k(yo+k3)
        s,sdot,v,vdot = yo + 1./6.*(
            k1 + 2.*k2 + 2.*k3 + k4
        )
    return map(np.array,(S,Sdot,V,Vdot,Ts,Tv,Us,Uv))

###########################
# calculate sx trajectory #
#     (using RK4)         #
###########################
#
# returns an 8 element list containing:
# 
# * s coordinates
# * s velocities
# * x coordinates
# * x velocities
# * s kinetic energy
# * x kinetic energy
# * s potential energy
# * x potential energy
#
def calculate_sx_trajectory(
        dt, # time step in vibrational periods (1/omegavo)
        s, # initial s-coord (ang)
        sdot, # initial s-coord velocity (ang . omegavo)
        x, # initial v-coord (ang)
        xdot, # initial v-coord velocity (ang . omegavo)
        smin, smax, # terminate simulation if s < smin or s > smax
):
    S = []
    Sdot = []
    X = []
    Xdot = []
    Ts = []
    Tx = []
    Us = []
    Ux = []
    while s > smin and s < smax:
        ts,tv,tx,us,uv,ux = energies(s,sdot,0,0,x,xdot)
        S.append(s)
        Sdot.append(sdot)
        X.append(x)
        Xdot.append(xdot)
        Ts.append(ts)
        Tx.append(tx)
        Us.append(us)
        Ux.append(ux)
        def k(y):
            return dt * np.array(
                [
                    y[1],
                    fs(
                        y[0],y[1],0,0,y[2]
                    ),
                    y[3],
                    fx(y[0],y[2])
                ]
            )
        yo = np.array(
            [s,sdot,x,xdot]
        )    
        k1 = k(yo)
        k2 = k(yo+k1/2.)
        k3 = k(yo+k2/2.)
        k4 = k(yo+k3)
        s,sdot,x,xdot = yo + 1./6.*(
            k1 + 2.*k2 + 2.*k3 + k4
        )
    return map(np.array,(S,Sdot,X,Xdot,Ts,Tx,Us,Ux))
